// Fractal.ts
/**
 * Base game map script - Produces widely varied continents.
 * @packageDocumentation
 */
console.log("Generating using script fractal.ts");
import { assignStartPositions, chooseStartSectors } from '/base-standard/maps/assign-starting-plots.js';
import { addMountains, addHills, buildRainfallMap, generateLakes } from '/base-standard/maps/elevation-terrain-generator.js';
import { addFeatures, designateBiomes } from '/base-standard/maps/feature-biome-generator.js';
import * as globals from '/base-standard/maps/map-globals.js';
import * as utilities from '/base-standard/maps/map-utilities.js';
import { addNaturalWonders } from '/base-standard/maps/natural-wonder-generator.js';
import { generateResources } from '/base-standard/maps/resource-generator.js';
import { addVolcanoes } from '/base-standard/maps/volcano-generator.js';
import { assignAdvancedStartRegions } from '/base-standard/maps/assign-advanced-start-region.js';
import { generateDiscoveries } from '/base-standard/maps/discovery-generator.js';
import { generateSnow, dumpPermanentSnow } from '/base-standard/maps/snow-generator.js';
import { dumpStartSectors, dumpContinents, dumpTerrain, dumpElevation, dumpRainfall, dumpBiomes, dumpFeatures, dumpResources, dumpNoisePredicate } from '/base-standard/maps/map-debug-helpers.js';
function requestMapData(initParams) {
    console.log(initParams.width);
    console.log(initParams.height);
    console.log(initParams.topLatitude);
    console.log(initParams.bottomLatitude);
    console.log(initParams.wrapX);
    console.log(initParams.wrapY);
    console.log(initParams.mapSize);
    engine.call("SetMapInitData", initParams);
}
function generateMap() {
    console.log("Generating a map!");
    console.log(`Age - ${GameInfo.Ages.lookup(Game.age).AgeType}`);
    let iWidth = GameplayMap.getGridWidth();
    let iHeight = GameplayMap.getGridHeight();
    let uiMapSize = GameplayMap.getMapSize();
    let startPositions = [];
    let mapInfo = GameInfo.Maps.lookup(uiMapSize);
    if (mapInfo == null)
        return;
    // Establish continent boundaries
    let iOceanWaterColumns = globals.g_OceanWaterColumns + 4;
    let westContinent = {
        west: iOceanWaterColumns / 2,
        east: (iWidth / 2) - (iOceanWaterColumns / 2),
        south: globals.g_PolarWaterRows + 2,
        north: iHeight - (globals.g_PolarWaterRows + 2),
        continent: 0
    };
    let eastContinent = {
        west: (iWidth / 2) + (iOceanWaterColumns / 2),
        east: iWidth - (iOceanWaterColumns / 2),
        south: globals.g_PolarWaterRows + 2,
        north: iHeight - (globals.g_PolarWaterRows + 2),
        continent: 0
    };

    let startSectors;
    let iNumPlayers1 = mapInfo.PlayersLandmass1;
    let iNumPlayers2 = mapInfo.PlayersLandmass2;
    let iNumNaturalWonders = Math.round(mapInfo.NumNaturalWonders * 1);
    let iTilesPerLake = mapInfo.LakeGenerationFrequency;
    let iStartSectorRows = mapInfo.StartSectorRows;
    let iStartSectorCols = mapInfo.StartSectorCols;
    let iRandom = TerrainBuilder.getRandomNumber(2, "East or West");
    console.log("Random Hemisphere: " + iRandom);
    if (iRandom == 1) {
        let iNum1 = iNumPlayers1;
        let iNum2 = iNumPlayers2;
        iNumPlayers1 = iNum2;
        iNumPlayers2 = iNum1;
    }
    let seed = GameplayMap.getRandomSeed(); // can use any seed you want for different noises
    let bHumanNearEquator = utilities.needHumanNearEquator();
    startSectors = chooseStartSectors(iNumPlayers1, iNumPlayers2, iStartSectorRows, iStartSectorCols, bHumanNearEquator);
    createLandmasses(iWidth, iHeight, westContinent, eastContinent, iStartSectorRows, iStartSectorCols, startSectors, seed);
    TerrainBuilder.validateAndFixTerrain();
    expandCoastsPlus(westContinent.west, westContinent.east, iHeight);
    expandCoastsPlus(eastContinent.west, eastContinent.east, iHeight);
    expandCoastsPlus(0, westContinent.west - globals.g_OceanWaterColumns, iHeight);
    expandCoastsPlus(westContinent.east + globals.g_OceanWaterColumns, eastContinent.west - globals.g_OceanWaterColumns, iHeight);
    expandCoastsPlus(eastContinent.east + globals.g_OceanWaterColumns, 0, iHeight);
    utilities.adjustOceanPlotTags(iNumPlayers1 > iNumPlayers2);
    AreaBuilder.recalculateAreas();
    TerrainBuilder.stampContinents();
    addMountains(iWidth, iHeight);
    addVolcanoes(iWidth, iHeight);
    generateLakes(iWidth, iHeight, iTilesPerLake);
    AreaBuilder.recalculateAreas();
    TerrainBuilder.buildElevation();
    addHills(iWidth, iHeight);
    buildRainfallMap(iWidth, iHeight);
    TerrainBuilder.modelRivers(5, 15, globals.g_NavigableRiverTerrain);
    TerrainBuilder.validateAndFixTerrain();
    TerrainBuilder.defineNamedRivers();
    designateBiomes(iWidth, iHeight);
    addNaturalWonders(iWidth, iHeight, iNumNaturalWonders);
    TerrainBuilder.addFloodplains(4, 10);
    addFeatures(iWidth, iHeight);
    TerrainBuilder.validateAndFixTerrain();
    for (let iY = 0; iY < iHeight; iY++) {
        for (let iX = 0; iX < iWidth; iX++) {
            let terrain = GameplayMap.getTerrainType(iX, iY);
            if (terrain == globals.g_CoastTerrain) {
                TerrainBuilder.setPlotTag(iX, iY, PlotTags.PLOT_TAG_WATER);
                if (iNumPlayers1 > iNumPlayers2) {
                    if (iX < westContinent.west - 2 || iX > westContinent.east + 2) {
                        //console.log("Islands on the Coast: " + iX + ", " + iY)
                        TerrainBuilder.addPlotTag(iX, iY, PlotTags.PLOT_TAG_EAST_WATER);
                    }
                    else {
                        //console.log("Main Coast: " + iX + ", " + iY)
                        TerrainBuilder.addPlotTag(iX, iY, PlotTags.PLOT_TAG_WEST_WATER);
                    }
                }
                else {
                    if (iX > eastContinent.east + 2 || iX < eastContinent.west - 2) {
                        //console.log("Islands on the Coast2: " + iX + ", " + iY)
                        TerrainBuilder.addPlotTag(iX, iY, PlotTags.PLOT_TAG_WEST_WATER);
                    }
                    else {
                        //console.log("Main Coast2: " + iX + ", " + iY)
                        TerrainBuilder.addPlotTag(iX, iY, PlotTags.PLOT_TAG_EAST_WATER);
                    }
                }
            }
        }
    }
    AreaBuilder.recalculateAreas();
    TerrainBuilder.storeWaterData();
    generateSnow(iWidth, iHeight);
    dumpStartSectors(startSectors);
    dumpContinents(iWidth, iHeight);
    dumpTerrain(iWidth, iHeight);
    dumpElevation(iWidth, iHeight);
    dumpRainfall(iWidth, iHeight);
    dumpBiomes(iWidth, iHeight);
    dumpFeatures(iWidth, iHeight);
    dumpPermanentSnow(iWidth, iHeight);
    generateResources(iWidth, iHeight, westContinent, eastContinent, iNumPlayers1, iNumPlayers2);
    startPositions = assignStartPositions(iNumPlayers1, iNumPlayers2, westContinent, eastContinent, iStartSectorRows, iStartSectorCols, startSectors);
    generateDiscoveries(iWidth, iHeight, startPositions);
    dumpResources(iWidth, iHeight);
    FertilityBuilder.recalculate(); // Must be after features are added.
    let avgDistanceBetweenPoints = 3;
    let normalizedRangeSmoothing = 2;
    let poisson = TerrainBuilder.generatePoissonMap(seed, avgDistanceBetweenPoints, normalizedRangeSmoothing);
    let poissonPred = (val) => {
        return val >= 1 ? "*" : " ";
    };
    dumpNoisePredicate(iWidth, iHeight, poisson, poissonPred);
    assignAdvancedStartRegions();
}
// Register listeners.
engine.on('RequestMapInitData', requestMapData);
engine.on('GenerateMap', generateMap);
console.log("Loaded fractal.ts");
function createLandmasses(iWidth, iHeight, westContinent, eastContinent, iStartSectorRows, iStartSectorCols, startSectors, seed) {
    FractalBuilder.create(globals.g_LandmassFractal, iWidth, iHeight, 3, 0);
    let waterPercentContinents = globals.g_WaterPercent + 5; // Total of 31%
    let iWaterHeight = FractalBuilder.getHeightFromPercent(globals.g_LandmassFractal, waterPercentContinents);
    let iBuffer = Math.floor(iHeight / 13.5);
    let iBuffer2 = Math.floor(iWidth / 21.0);
    let iBufferIslands = Math.floor(iWidth / 24.0);
    let eastCenterX = eastContinent.west + (eastContinent.east - eastContinent.west) / 2.0;
    let eastCenterY = eastContinent.south + (eastContinent.north - eastContinent.south) / 2.0;
    let eastRadius = (eastContinent.east - eastContinent.west) / 2.3;
    let westCenterX = westContinent.west + (westContinent.east - westContinent.west) / 2.0;
    let westCenterY = westContinent.south + (westContinent.north - westContinent.south) / 2.0;
    let westRadius = (westContinent.east - westContinent.west) / 2.3;

    for (let iY = 0; iY < iHeight; iY++) {
        for (let iX = 0; iX < iWidth; iX++) {
            let terrain = globals.g_FlatTerrain;
            let iRandom = 0;
            let iRandom2 = 0;
            // Initialize plot tag
            TerrainBuilder.setPlotTag(iX, iY, PlotTags.PLOT_TAG_NONE);
            //  Must be water if at the poles
            if (iY < westContinent.south + iRandom || iY >= westContinent.north - iRandom) {
                terrain = globals.g_OceanTerrain;
            }
            // Of if between the continents
            else if (iX < westContinent.west + iRandom2 || iX >= eastContinent.east - iRandom2 ||
                (iX >= westContinent.east - iRandom2 && iX < eastContinent.west + iRandom2)) {
                terrain = globals.g_OceanTerrain;
            }
            else {
                // Get the value from the fractal
                let iPlotHeight = utilities.getHeightAdjustingForStartSector(iX, iY, iWaterHeight, globals.g_FractalWeight, 0.0 /*CenterWeight*/, globals.g_StartSectorWeight, westContinent, eastContinent, iStartSectorRows, iStartSectorCols, startSectors);
                if (iX >= iWidth / 2) {
                    let difX = Math.abs(iX - eastCenterX);
                    let difY = Math.abs(iY - eastCenterY) * 0.5;
                    let difXY = difX * difX + difY * difY;
                    let difNorm = Math.pow(difXY / eastRadius, 1.3);
                    iPlotHeight -= difNorm * 1.25;
                } else {
                    let difX = Math.abs(iX - westCenterX);
                    let difY = Math.abs(iY - westCenterY) * 0.5;
                    let difXY = difX * difX + difY * difY;
                    let difNorm = Math.pow(difXY / westRadius, 1.3);
                    iPlotHeight -= difNorm * 1.25;
                }


                // Finally see whether or not this stays as Land or has too low a score and drops back to water
                if (iPlotHeight < iWaterHeight * (globals.g_FractalWeight + globals.g_StartSectorWeight)) {
                    terrain = globals.g_OceanTerrain;
                }
            }
            setTerrain(iX, iY, terrain, iWidth / 2);
        }
    }

    // Randomly extend land to remove fractal patterns
    for (let test = 0; test < 30; test++) {
        for (let iY = 0; iY < iHeight; iY++) {
            for (let iX = 0; iX < iWidth; iX++) {
                if (isLandSpotValid(iX, iY, iWidth, seed)) {
                    let count = adjacentLandCount(iX, iY);

                    if (count == 1 && TerrainBuilder.getRandomNumber(256, "Shallow Water Scater Scatter1") == 0) {
                        setTerrain(iX, iY, globals.g_FlatTerrain, iWidth / 2);
                    }
                    if (count == 2 && TerrainBuilder.getRandomNumber(48, "Shallow Water Scater Scatter2") == 0) {
                        setTerrain(iX, iY, globals.g_FlatTerrain, iWidth / 2);
                    }
                    if (count == 3 && TerrainBuilder.getRandomNumber(16, "Shallow Water Scater Scatter3") == 0) {
                        setTerrain(iX, iY, globals.g_FlatTerrain, iWidth / 2);
                    }
                    if (count == 4 && TerrainBuilder.getRandomNumber(24, "Shallow Water Scater Scatter4") == 0) {
                        setTerrain(iX, iY, globals.g_FlatTerrain, iWidth / 2);
                    }
                    if (count == 5 && TerrainBuilder.getRandomNumber(8, "Shallow Water Scater Scatter5") == 0) {
                        setTerrain(iX, iY, globals.g_FlatTerrain, iWidth / 2);
                    }
                    if (count == 6 && TerrainBuilder.getRandomNumber(4, "Shallow Water Scater Scatter6") == 0) {
                        setTerrain(iX, iY, globals.g_FlatTerrain, iWidth / 2);
                    }
                }
            }
        }
    }

    let totalTiles = iWidth * iHeight;
    let islandPlacementCount = totalTiles / 60;
    for (let test = 0; test < islandPlacementCount; test++) {
        let randStartSize = 4 + TerrainBuilder.getRandomNumber(6, "rand start size");
        let randExtraSize = 4 + TerrainBuilder.getRandomNumber(12, "rand extra size");
        let randSizeRatio = TerrainBuilder.getRandomNumber(randExtraSize, "rand XY ratio");
        let sizeX = randStartSize + randSizeRatio;
        let sizeY = randStartSize + (randExtraSize - randSizeRatio);
        let posX = -sizeX + TerrainBuilder.getRandomNumber(sizeX + iWidth, "rand pos X");
        let posY = -sizeY + TerrainBuilder.getRandomNumber(sizeY + iHeight, "rand pos Y");

        placeIslandsInRegion(posX, posY, iWidth, iHeight, sizeX, sizeY, seed, iWidth / 2);
    }

    for (let iY = 0; iY < iHeight; iY++) {
        for (let iX = 0; iX < iWidth; iX++) {
            let iRandom = TerrainBuilder.getRandomNumber(iBuffer, "Random Top/Bottom Edges");
            if (iY < 0 + iRandom + 1 || iY >= iHeight - (iRandom + 1)) {
                setTerrain(iX, iY, globals.g_OceanTerrain, iWidth / 2);
            }
        }
    }
}

export function placeIslandsInRegion(iX, iY, iWidth, iHeight, sizeX, sizeY, seed, distantLandEdge) {
    let weightSum = 0;
    let weightCount = 0;
    let validArray = new Array(sizeX * sizeY);
    for (let y = 0; y < sizeY; y++) {
        for (let x = 0; x < sizeX; x++) {
            let isSpotValid = true;
            let X = x + iX;
            let Y = y + iY;
            if (X < 0) { X += iWidth; }
            if (X >= iWidth) { X = X % iWidth; }
            if (Y < 0 || Y >= iHeight) { isSpotValid = false; }
            if (!isLandSpotValid(X, Y, iWidth, seed)) { isSpotValid = false; }


            if (!isSpotValid) {
                validArray[x + y * sizeX] = false;
                weightCount++;
                if (GameplayMap.hasPlotTag(X, Y, PlotTags.PLOT_TAG_ISLAND)) {
                    weightSum += 1.0;
                }
            } else {
                validArray[x + y * sizeX] = isOnlyWaterAround(X, Y);
            }
        }
    }
    let islandRatio = weightSum / parseFloat(weightCount);
    if (islandRatio > 0.15) { return; }

    FractalBuilder.create(globals.g_LandmassFractal, iWidth, iHeight, 4 + TerrainBuilder.getRandomNumber(2, "fractal offset"), 0);
    let waterPercentIslands = 68 + TerrainBuilder.getRandomNumber(20, "extra island water percent");
    let iWaterHeightIsland = FractalBuilder.getHeightFromPercent(globals.g_LandmassFractal, waterPercentIslands);

    for (let y = 0; y < sizeY; y++) {
        for (let x = 0; x < sizeX; x++) {
            if (!validArray[x + y * sizeX]) {
                continue;
            }

            let X = x + iX;
            let Y = y + iY;
            if (X < 0) { X += iWidth; }
            if (X >= iWidth) { X = X % iWidth; }

            if (Y >= 0 && Y < iHeight) {
                let iPlotHeight = FractalBuilder.getHeight(globals.g_LandmassFractal, X, Y);

                let difX = Math.abs(x - sizeX / 2.0) / parseFloat(sizeX / 2.0);
                let difY = Math.abs(y - sizeY / 2.0) / parseFloat(sizeY / 2.0);
                let difXY = difX * difX + difY * difY;
                let difNorm = Math.pow(difXY, 1.0);
                iPlotHeight -= difNorm * 0.4;

                if (iPlotHeight > iWaterHeightIsland) {
                    //TerrainBuilder.setTerrainType(X, Y, globals.g_FlatTerrain);
                    //TerrainBuilder.addPlotTag(X, Y, PlotTags.PLOT_TAG_ISLAND);
                    setTerrain(X, Y, globals.g_FlatTerrain, distantLandEdge);
                }
            }
        }
    }
}


export function expandCoastsPlus(iWest, iEast, iHeight) {
    for (let iY = 0; iY < iHeight; iY++) {
        for (let iX = iWest; iX < iEast; iX++) {
            let terrain = GameplayMap.getTerrainType(iX, iY);
            if (terrain == globals.g_OceanTerrain) {
                if (GameplayMap.isAdjacentToShallowWater(iX, iY) && TerrainBuilder.getRandomNumber(2, "Shallow Water Scater Scatter") == 0) {
                    TerrainBuilder.setTerrainType(iX, iY, globals.g_CoastTerrain);
                }
            }
        }
    }
}

export function adjacentLandCount(iX, iY) {
    let count = 0;
    if (!GameplayMap.hasPlotTag(iX, iY, PlotTags.PLOT_TAG_LANDMASS)) {
        for (let iDirection = 0; iDirection < DirectionTypes.NUM_DIRECTION_TYPES; iDirection++) {
            let iIndex = GameplayMap.getIndexFromXY(iX, iY);
            let iLocation = GameplayMap.getLocationFromIndex(iIndex);
            let iAdjacentX = GameplayMap.getAdjacentPlotLocation(iLocation, iDirection).x;
            let iAdjacentY = GameplayMap.getAdjacentPlotLocation(iLocation, iDirection).y;
            if (GameplayMap.hasPlotTag(iAdjacentX, iAdjacentY, PlotTags.PLOT_TAG_LANDMASS)) {
                count++;
            }
        }
    }
    return count;
}

export function isLandSpotValid(iX, iY, iWidth, seed) {

    let seedAtHeight = seed + iY;
    let hash = seedAtHeight;
    let rand = Math.max(0, hash % 4);
    let leftRand = rand < 1 ? 1 : 0;
    let rightRand = rand >= 3 ? 1 : 0;

    if (iX < 2 + rightRand || iX >= (iWidth - 1) - leftRand) {
        return false;
    }
    let iWidth2 = iWidth / 2;
    if (iX >= (iWidth2 - 1) - leftRand && iX < (iWidth2 + 2) + rightRand) {
        return false;
    }
    return true;
}

export function isOnlyWaterAround(iX, iY, plotTag = PlotTags.PLOT_TAG_LANDMASS) {
    let count = 0;
    if (!GameplayMap.hasPlotTag(iX, iY, plotTag)) {
        for (let iDirection = 0; iDirection < DirectionTypes.NUM_DIRECTION_TYPES; iDirection++) {
            let iIndex = GameplayMap.getIndexFromXY(iX, iY);
            let iLocation = GameplayMap.getLocationFromIndex(iIndex);
            let iAdjacentX = GameplayMap.getAdjacentPlotLocation(iLocation, iDirection).x;
            let iAdjacentY = GameplayMap.getAdjacentPlotLocation(iLocation, iDirection).y;
            if (GameplayMap.hasPlotTag(iAdjacentX, iAdjacentY, PlotTags.PLOT_TAG_WATER)) {
                count++;
            }
        }
    }
    return count == 6;
}

export function setTerrain(iX, iY, terrain, eastWest) {
    if (terrain != globals.g_OceanTerrain && terrain != globals.g_CoastTerrain) {
        utilities.addLandmassPlotTags(iX, iY, eastWest);
    }
    else {
        utilities.addWaterPlotTags(iX, iY, eastWest);
    }
    TerrainBuilder.setTerrainType(iX, iY, terrain);
}

//# sourceMappingURL=file:///base-standard/maps/fractal.js.map
